/***************************************************************************
 *  Copyright (C) 2009 Nippon Telegraph and Telephone Corporation
 ***************************************************************************/

CREATE LANGUAGE plpgsql;

BEGIN;

CREATE SCHEMA statsinfo;
SET search_path = statsinfo;
SET client_encoding TO UTF8;

/**************************************
 * Snapshot ID table
 *************************************/
CREATE TABLE t_snapshot_id (
	snapshot_id      bigserial,
	snapshot_date    timestamp with time zone,
	dbid             bigint,
	host_id          bigint,
	level            smallint,
	description      text,
	primary key (snapshot_id)
);

CREATE INDEX i_snapshot_id_date ON t_snapshot_id(snapshot_date);


/**************************************
 * DB connection information table
 *************************************/
CREATE TABLE t_dbconn (
	dbid                bigserial,
	dbname              text,
	hostname            text,
	port                integer,
	username            text,
	password            text,
	is_snapshot_target  boolean,
	primary key (dbid),
	unique (dbname, hostname, port)
);


/**************************************
 * Host ID table
 *************************************/
CREATE TABLE t_host_id (
	host_id             bigserial,
    hostname            text,
    port                integer,
    primary key (host_id)
);


/**************************************
 * Host basic information table
 *************************************/
CREATE TABLE t_hostinfo (
	host_snapshot_id   bigserial,
	host_snapshot_date timestamp with time zone,
	host_id            bigint,
	kernel_version     text,
	processors         bigint,
	mem_total          bigint,
	swap_total         bigint,
	pg_version         text,
	maxwritten_clean   bigint,
    primary key (host_snapshot_id)
);

CREATE INDEX i_hostinfo_snapshot_date ON t_hostinfo (host_snapshot_date);

/**************************************
 * Host long transaction infomation table
 *************************************/
CREATE TABLE t_hostactivity (
	host_snapshot_id    bigint,
	dbname              text,
	procpid             int,
	usename             text,
	client_addr         inet,
	waiting             boolean,
	duration            interval(0),
	check_time          timestamp(0) with time zone,
	backend_start       timestamp(0) with time zone,
	current_query       text,
	FOREIGN KEY (host_snapshot_id) REFERENCES statsinfo.t_hostinfo (host_snapshot_id) ON DELETE CASCADE
);

CREATE INDEX i_hostactivity_id ON t_hostactivity (host_snapshot_id);


/**************************************
 * Host device usage information table
 *************************************/
CREATE TABLE t_fs_size (
	host_snapshot_id       bigint,
	tblspc_oid             oid,
	name                   text,
	device_id              int,
	device_name            text,
	location               text,
	used                   bigint,
	avail                  bigint,
	total                  bigint,
	FOREIGN KEY (host_snapshot_id) REFERENCES statsinfo.t_hostinfo (host_snapshot_id) ON DELETE CASCADE
);

CREATE INDEX i_t_fs_size_id ON t_fs_size (host_snapshot_id);


/**************************************
 * DB basic information table
 *************************************/
CREATE TABLE t_dbinfo (
	snapshot_id      bigint,
	dbid             bigint,
	dbsize           bigint,
	conn_number      integer,
	datfrozenxid_age xid,
	num_user_rel     bigint,
	db_xact_commit   bigint,
	db_xact_rollback bigint,
	db_blks_read     bigint,
	db_blks_hit      bigint,
	db_tup_returned  bigint,
	db_tup_fetched   bigint,
	db_tup_inserted  bigint,
	db_tup_updated   bigint,
	db_tup_deleted   bigint,
    primary key (snapshot_id, dbid),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);

/**************************************
 * DB configuration information table
 *************************************/
CREATE TABLE t_dbsetting (
	snapshot_id    bigint,
	dbid           bigint,
	name           text,
	setting        text,
	source         text,
    primary key (snapshot_id, dbid, name),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);


/**************************************
 * Table density information table
 *************************************/
CREATE TABLE t_tablefill (
	snapshot_id    bigint,
	dbid           bigint,
	schemaname     text,
	relname        text,
	reltuples      real,
	logical_pages  bigint,
	physical_pages bigint,
	tratio         int,
    primary key (snapshot_id, dbid, schemaname, relname),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);


/******************************************
 * Low correlation column information table
 ******************************************/
CREATE TABLE t_columncorr (
	snapshot_id   bigint,
	dbid          bigint,
	schemaname    text,
	relname       text,
	column_name   text,
	correlation   real,
    primary key (snapshot_id, dbid, schemaname, relname),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);


/**************************************
 * Table statistical information table
 *************************************/
CREATE TABLE t_tableinfo (
	snapshot_id    bigint,
	dbid           bigint,
	schemaname     text,
	table_oid      oid,
	tablename      text,
	relpages       bigint,
	reltuples      real,
	n_tup_ins      bigint,
	n_tup_upd      bigint,
	n_tup_del      bigint,
	n_tup_hot_upd  bigint,
	seq_scan       bigint,
	idx_scan       bigint,
	seq_tup_read   bigint,
	idx_tup_fetch  bigint,
	heap_blks_hit  bigint,
	idx_blks_hit   bigint,
	toast_blks_hit bigint,
	tidx_blks_hit  bigint,
	heap_blks_read bigint,
	idx_blks_read  bigint,
	toast_blks_read bigint,
	tidx_blks_read bigint,
	n_live_tuple   bigint,
	n_dead_tuple   bigint,
	relation_size  bigint,
	last_vacuum      timestamp with time zone,
	last_autovacuum  timestamp with time zone,
	last_analyze     timestamp with time zone,
	last_autoanalyze timestamp with time zone,
    primary key (snapshot_id, dbid, schemaname, table_oid),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);


/**************************************
 * Index statistical information table
 *************************************/
CREATE TABLE t_indexinfo (
	snapshot_id    bigint,
	dbid           bigint,
	schemaname     text,
	index_oid      oid,
	indexname      text,
	tablename      text,
	indexdef       text,
	cluster_key    boolean,
	index_size     bigint,
	idx_scan       bigint,
	bmp_tup_fetch   bigint,
	idx_tup_fetch  bigint,
	idx_blks_read  bigint,
	idx_blks_hit   bigint,
    primary key (snapshot_id, dbid, schemaname, index_oid),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE

);

/**************************************
 * Column statistical information table
 *************************************/
CREATE TABLE t_columninfo (
	snapshot_id    bigint,
	dbid           bigint,
	schemaname     text,
	tablename      text,
	column_name    text,
	n_distinct     real,
	correlation    real,
    primary key (snapshot_id, dbid, schemaname, tablename, column_name),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);


/**************************************
 * autovacuum information table
 *************************************/
CREATE TABLE t_autovacuuminfo (
	snapshot_id                bigint,
	dbid                       bigint,
	autovacuum_act_time        timestamp with time zone,
	autovacuum_table           text,
	autovacuum_num_page_remove bigint,
	autovacuum_num_page_remain bigint,
	autovacuum_num_tup_remove  bigint,
	autovacuum_num_tup_remain  bigint,
	autovacuum_duration        real,
    primary key (snapshot_id, dbid, autovacuum_act_time, autovacuum_table, autovacuum_num_page_remove,
                 autovacuum_num_page_remain, autovacuum_num_tup_remove, autovacuum_num_tup_remain, autovacuum_duration),
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);


/**************************************
 * checkpoint information table
 *************************************/
CREATE TABLE t_checkpointinfo (
	host_snapshot_id          bigint,
	checkpoint_act_time       timestamp with time zone,
	checkpoint_start_trig     text,
	checkpoint_num_buffers    bigint,
	checkpoint_create_wal     bigint,
	checkpoint_delete_wal     bigint,
	checkpoint_recycle_wal    bigint,
	checkpoint_write_duration real,
	checkpoint_sync_duration  real,
	checkpoint_total_duration real,
	FOREIGN KEY (host_snapshot_id) REFERENCES statsinfo.t_hostinfo (host_snapshot_id) ON DELETE CASCADE
);


/**************************************
 * pg_stat_statements information table
 *************************************/
CREATE TABLE t_statstatements (
	host_snapshot_id	bigint,
	usename				name,
	datname				name,
	query				text,
	calls				bigint,
	total_time			double precision,
	rows				bigint,
	FOREIGN KEY (host_snapshot_id) REFERENCES statsinfo.t_hostinfo (host_snapshot_id) ON DELETE CASCADE
);


/**************************************
 * pg_stat_user_functions information table
 *************************************/
CREATE TABLE t_statfunc (
	snapshot_id			bigint,
	dbid				bigint,
	schemaname			name,
	funcname			name,
	calls				bigint,
	total_time			bigint,
	self_time			bigint,
	FOREIGN KEY (snapshot_id) REFERENCES statsinfo.t_snapshot_id (snapshot_id) ON DELETE CASCADE
);



/* t_dbconn table has user-password, so we performe encode password on 
 * before trigger against insert or update.
 */
CREATE OR REPLACE FUNCTION check_new_host() RETURNS trigger AS $$
	DECLARE
		cnt integer;
	BEGIN
		IF (NEW.password IS NULL) THEN
			NEW.password = '';
		END IF;
		IF (TG_OP = 'INSERT') THEN
			SELECT COUNT(*) INTO STRICT cnt FROM statsinfo.t_host_id WHERE hostname = NEW.hostname AND port = NEW.port;
			IF cnt = 0 THEN
				INSERT INTO statsinfo.t_host_id VALUES(default, NEW.hostname, NEW.port);
			END IF;
			NEW.password = encode(pgp_sym_encrypt(NEW.password, statsinfo.get_key())::bytea, 'hex');
			RETURN NEW;
		ELSIF (TG_OP = 'UPDATE') THEN
			IF(NEW.password = OLD.password) THEN
				RETURN NEW;
			ELSE
				NEW.password = encode(pgp_sym_encrypt(NEW.password, statsinfo.get_key())::bytea, 'hex');
				RETURN NEW;
			END IF;
		END IF;			
	END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER trig_dbconn BEFORE INSERT OR UPDATE ON statsinfo.t_dbconn FOR EACH ROW EXECUTE PROCEDURE statsinfo.check_new_host();

CREATE OR REPLACE FUNCTION get_key() 
RETURNS text
AS '$libdir/stats_info','get_key'
LANGUAGE C STRICT;


/* getting snapshot (from all flag-is-on DB) */
CREATE OR REPLACE FUNCTION get_stats() 
RETURNS boolean
AS '$libdir/stats_info','get_stats_no_args'
LANGUAGE C STRICT;

/* getting snapshot（from dbid） */
CREATE OR REPLACE FUNCTION get_stats(bigint) 
RETURNS boolean
AS '$libdir/stats_info','get_stats_by_dbid'
LANGUAGE C STRICT;

/* getting snapshot（from all falg-is-on DB with description） */
CREATE OR REPLACE FUNCTION get_stats(text) 
RETURNS boolean
AS '$libdir/stats_info','get_stats_with_description'
LANGUAGE C STRICT;

/* getting snapshot（from dbid with description） */
CREATE OR REPLACE FUNCTION get_stats(bigint, text) 
RETURNS boolean
AS '$libdir/stats_info','get_stats_by_dbid_with_description'
LANGUAGE C STRICT;

/* displayng snapshot information */
CREATE VIEW snapshot_history AS 
SELECT
    ss.snapshot_id as snapshot_id,
    ss.snapshot_date as snapshot_date,
    dc.hostname as hostname,
    dc.dbname as dbname,
    dc.port as port,
    ss.level as level,
    ss.description as description
FROM
    statsinfo.t_snapshot_id ss,
    statsinfo.t_dbconn dc
WHERE
    ss.dbid = dc.dbid
ORDER BY
    ss.snapshot_id;

/* deletion snapshot 
 *
 *  We must not delete information only based on snapshot_id.
 *  Because each snapshots share the host-information. (ex. filterd-checkpotint-info)
 *  Thus, we check delible host_sbaphot_id based on timestamp in t_snapshot_id.
 */

CREATE OR REPLACE FUNCTION delete_stats(id bigint)
RETURNS boolean
AS $$
DECLARE
    check int;
BEGIN

    -- other DBs (on the same host) needs hostinfo that be scheduled to go on delete ?
    SELECT 1 INTO check FROM statsinfo.t_snapshot_id
      WHERE snapshot_id < id AND host_id = (SELECT host_id FROM statsinfo.t_snapshot_id WHERE snapshot_id = id) LIMIT 1;

    IF  check IS NULL THEN
    -- we can delete all relative snapshots.
    DELETE FROM statsinfo.t_hostinfo  WHERE host_snapshot_id IN
     (SELECT host_snapshot_id FROM statsinfo.t_hostinfo h ,
       (SELECT snapshot_date, host_id FROM statsinfo.t_snapshot_id WHERE snapshot_id = id) s
      WHERE h.host_snapshot_date <= s.snapshot_date AND h.host_id = s.host_id
     );
    END IF;

    DELETE FROM statsinfo.t_snapshot_id WHERE snapshot_id = id;

RETURN 't';
END;
$$ LANGUAGE plpgsql;

/*
 * deletion autovacuum-log.done file.
 * This function is called by del_avlog_trigger on t_snapshot_id.
 * The trigger will be performed after commit of "INSERT INTO t_snapshot_id".
 */
CREATE OR REPLACE FUNCTION del_avlog() RETURNS TRIGGER AS
$$
DECLARE
  conn_info text;
  target_db text;
  query     text;
  result   boolean;
BEGIN

  -- get databas name
  SELECT dbname INTO target_db FROM statsinfo.t_dbconn WHERE dbid = NEW.dbid;

  -- get connection info
  SELECT 'dbname=' || dbname || ' host=' || hostname || ' port=' || port || ' user=' || username || ' password=' || pgp_sym_decrypt(decode(password, 'hex'), get_key()) INTO conn_info FROM t_dbconn
  WHERE dbid = NEW.dbid;

  PERFORM dblink_connect('del_av', conn_info);


  SELECT res INTO result FROM dblink('del_av','SELECT pg_file_unlink (''pg_statslog/' || target_db || '-autovacuum.log.done'')') AS t  (res boolean);

  FOR query IN SELECT 'SELECT pg_file_unlink (''pg_statslog/' || datname || '-autovacuum.log'')'
    FROM dblink('del_av','SELECT datname FROM pg_database') as t (datname name)
      WHERE datname NOT IN (SELECT dbname FROM t_dbconn) LOOP
    SELECT res INTO result FROM dblink('del_av', query) AS t (res boolean);
  END LOOP;

  PERFORM dblink_disconnect('del_av');

  RETURN NEW;

END;
$$ LANGUAGE plpgsql;


CREATE TRIGGER del_avlog_trigger AFTER INSERT ON statsinfo.t_snapshot_id
 FOR EACH ROW EXECUTE PROCEDURE del_avlog();

/* Set delay trigger */
UPDATE pg_catalog.pg_trigger
   SET tgdeferrable = true, tginitdeferred = true
 WHERE tgrelid = 'statsinfo.t_snapshot_id'::regclass
   AND tgname = 'del_avlog_trigger';

/*
 * deletion checkpoint-log.done file.
 * This function is called by del_cplog_trigger on t_hostinfo.
 * The trigger will be performed after commit of "INSERT INTO t_hostinfo".
 */
CREATE OR REPLACE FUNCTION del_cplog() RETURNS TRIGGER AS
$$
DECLARE
  conninfo text;
  result   text;
BEGIN

  -- try to connect target host. we use some connection info
  <<conn_check>>
  FOR conninfo IN
  SELECT 'dbname=' || c.dbname || ' host=' || c.hostname || ' port=' || c.port || ' user=' || c.username || ' password=' ||    pgp_sym_decrypt(decode(c.password, 'hex'), get_key())
  FROM (SELECT dbname, hostname, port, username, password FROM t_dbconn) c, t_hostinfo h, t_host_id i
  WHERE h.host_snapshot_id = NEW.host_snapshot_id  AND h.host_id = i.host_id AND i.hostname = c.hostname AND i.port = c.port LOOP

  -- connection success ?
  BEGIN
    PERFORM dblink_connect('del_cp',conninfo);
  -- Yes.
    EXIT conn_check;
  -- No. we try to connect with other conninfo.
  EXCEPTION WHEN OTHERS THEN
    CONTINUE;
  END;
  END LOOP;

  -- deletion checkpoint log.
  SELECT res INTO result FROM dblink('del_cp','SELECT pg_file_unlink(''pg_statslog/checkpoint.log.done'')') AS t (res boolean);

  PERFORM dblink_disconnect('del_cp');

  RETURN NEW;

END;
$$ LANGUAGE plpgsql;

CREATE TRIGGER del_cplog_trigger AFTER INSERT ON statsinfo.t_hostinfo
 FOR EACH ROW EXECUTE PROCEDURE del_cplog();

/* Set delay trigger */
UPDATE pg_catalog.pg_trigger
   SET tgdeferrable = true, tginitdeferred = true
 WHERE tgrelid = 'statsinfo.t_hostinfo'::regclass
   AND tgname = 'del_cplog_trigger';


REVOKE ALL ON t_snapshot_id FROM PUBLIC;
REVOKE ALL ON t_dbconn FROM PUBLIC;
REVOKE ALL ON t_host_id FROM PUBLIC;
REVOKE ALL ON t_dbinfo FROM PUBLIC;
REVOKE ALL ON t_dbsetting FROM PUBLIC;
REVOKE ALL ON t_tablefill FROM PUBLIC;
REVOKE ALL ON t_columncorr FROM PUBLIC;
REVOKE ALL ON t_tableinfo FROM PUBLIC;
REVOKE ALL ON t_indexinfo FROM PUBLIC;
REVOKE ALL ON t_columninfo FROM PUBLIC;
REVOKE ALL ON t_autovacuuminfo FROM PUBLIC;
REVOKE ALL ON t_checkpointinfo FROM PUBLIC;
REVOKE ALL ON t_statstatements FROM PUBLIC;
REVOKE ALL ON t_statfunc FROM PUBLIC;
REVOKE ALL ON t_hostactivity FROM PUBLIC;
REVOKE ALL ON t_fs_size FROM PUBLIC;
REVOKE ALL ON snapshot_history FROM PUBLIC;

REVOKE ALL ON FUNCTION get_key() FROM PUBLIC;
REVOKE ALL ON FUNCTION check_new_host() FROM PUBLIC;

REVOKE ALL ON FUNCTION get_stats() FROM PUBLIC;
REVOKE ALL ON FUNCTION get_stats(bigint) FROM PUBLIC;
REVOKE ALL ON FUNCTION get_stats(text) FROM PUBLIC;
REVOKE ALL ON FUNCTION get_stats(bigint, text) FROM PUBLIC;
REVOKE ALL ON FUNCTION delete_stats(bigint) FROM PUBLIC;
REVOKE ALL ON FUNCTION del_avlog() FROM PUBLIC;
REVOKE ALL ON FUNCTION del_cplog() FROM PUBLIC;

COMMIT;
